/*
	STM32F0
	WS2812B control with SPI & DMA, PLL in use, SysTick-driven animation
	gbm, 11'2015
*/
#define PLLMUL 10	// 40 MHz clock
#include "board.h"
#include "WS2812.h"
//========================================================================
void SystemInit(void)
{
	RCC->CFGR = RCC_CFGR_PLLMULV(PLLMUL);
	RCC->CR |= RCC_CR_PLLON;	// turn PLL on
}
//========================================================================
static const struct init_entry_ init_table[] = 
{
	{&FLASH->ACR, FLASH_ACR_PRFTBE | 1},	// enable prefetch, 1 wait state
	{&RCC->CFGR, RCC_CFGR_PLLMULV(PLLMUL) | RCC_CFGR_SW_PLL},	// switch to PLL clock
	// enable peripherals
	{&RCC->APB1ENR, RCC_APB1ENR_PWREN},
	{&RCC->APB2ENR, RCC_APB2ENR_SPI1EN},
	{&RCC->AHBENR, RCC_AHBENR_RSTVAL | RCC_AHBENR_GPIOAEN | RCC_AHBENR_GPIOBEN
		| RCC_AHBENR_GPIOCEN | RCC_AHBENR_DMAEN},
	// port setup - SPI1 MOSI at PA7 and PB5; SPI1 is fun 0 (default)
	{&GPIOA->OSPEEDR, BF2(7, GPIO_OSPEEDR_MED)},	// MOSI - medium speed
	{&GPIOA->MODER, GPIOA_MODER_SWD	| BF2(7, GPIO_MODER_AF)},	// MOSI as AF
	{&GPIOB->OSPEEDR, BF2(5, GPIO_OSPEEDR_MED)},	// MOSI - medium speed
	{&GPIOB->MODER, BF2(5, GPIO_MODER_AF)},	// MOSI as AF
	// SPI setup
	{(__IO32p)&SPI1->CR2, SPI_CR2_DSIZE(16) | SPI_CR2_TXDMAEN},
	{(__IO32p)&SPI1->CR1, SPI_CR1_SSM | SPI_CR1_SSI | SPI_CR1_SPE | SPI_CR1_BRDIV16 | SPI_CR1_MSTR},	// enable
	//SysTick setup
	{&SCB->SHP[1], 0x80c00000},		// PendSV lowest priority, Systick higher
	{&SysTick->LOAD, SYSCLK_FREQ / SYSTICK_FREQ - 1},
	{&SysTick->VAL, 0},
	{&SysTick->CTRL, SysTick_CTRL_CLKSOURCE_Msk | SysTick_CTRL_TICKINT_Msk
		| SysTick_CTRL_ENABLE_Msk},
	// Flash unlock
	{&FLASH->KEYR, FLASH_FKEY1},
	{&FLASH->KEYR, FLASH_FKEY2},
	// PVD setup
	{&PWR->CR, PWR_CR_PLSV(5) | PWR_CR_PVDE},	// 2.58 V, enable
	// interrupts and sleep
	{&EXTI->RTSR, EXTI_MR_PVD},	// PVD
	{&EXTI->IMR, EXTI_MR_PVD},	// PVD
	{&NVIC->ISER[0], 1 << PVD_IRQn},	// enable interrupts
	{0, 0}	// 3108 vs 3252
};
//========================================================================
extern void app_init(void);
void WEAK app_init(void) { }

int main(void)
{
	while (!(RCC->CR & RCC_CR_PLLRDY));	// wait for PLL lock
	writeregs(init_table);
	app_init();
	SCB->SCR = SCB_SCR_SLEEPONEXIT_Msk;	// sleep while not in handler
	__WFI();	// go to sleep
}
//========================================================================
// WS2812 reset pulse width
// (50 us -> 120 bits @2.4 MHz, 125 bits @2.5 MHz)
#define WS_RST_FRM	8 // no. of 16-bit frames for RESET pulse
//========================================================================
#define NBYTES	(NPIXELS * 3)
// no. of 16-bit frames for LED string control (9 frames per pixel pair)
#define NFRAMES	(WS_RST_FRM + (NPIXELS + 1) / 2 * 9)

void WS2812_start(struct wspix_ *ptr)
{
	static uint16_t enc_wsdata[NFRAMES];

	static const uint16_t encode[] = {
		04444, 04446, 04464, 04466, 04644, 04646, 04664, 04666,
		06444, 06446, 06464, 06466, 06644, 06646, 06664, 06666
	};	// bit-to triple encoding table - 4->12 bits
	
	if (DMA1->ISR & DMA_ISR_TCIF3 || DMA1_Channel3->CCR == 0)
	{
		uint8_t *wsptr = (uint8_t *)ptr;
		uint16_t *ep = enc_wsdata + WS_RST_FRM;	// skip reset frames
		
		for (uint32_t i = 0; i < NBYTES; i +=2)
		{
			uint32_t ev;
			// encode 2 bytes as 3 16-bit words
			ev = encode[*wsptr >> 4] << 12;
			ev |= encode[*wsptr ++ & 0xf];	// 24 bits in ev
			*ep ++ = ev >> 8;	// store 16 bits
			ev <<= 12;	// 8 bits left
			ev |= encode[*wsptr >> 4];	// 20 b total
			*ep ++ = ev >> 4;	// 4 bits left
			*ep ++ = ev << 12 | encode[*wsptr ++ & 0xf];	// 16 bits
		}
		// setup WS2812 data transfer
		DMA1->IFCR = DMA_IFCR_CGIF3;
		DMA1_Channel3->CCR = 0;	// disable
		DMA1_Channel3->CPAR = (uint32_t)&SPI1->DR;
		DMA1_Channel3->CMAR = (uint32_t)enc_wsdata;
		DMA1_Channel3->CNDTR = NFRAMES;
		// medium priority, increment memory adress, mem->periph, enable
		DMA1_Channel3->CCR = DMA_CCR_PL_0 | DMA_CCR_MSIZE16 | DMA_CCR_PSIZE16
			| DMA_CCR_MINC | DMA_CCR_DIR | DMA_CCR_EN;
		// clear buffer
		DMA1_Channel1->CCR = 0;	// disable
		DMA1_Channel1->CPAR = (uint32_t)enc_wsdata;	// constant zero
		DMA1_Channel1->CMAR = (uint32_t)ptr;	// 
		DMA1_Channel1->CNDTR = (NBYTES + 3) >> 2;
		// increment memory adress, mem->periph, enable
		DMA1_Channel1->CCR = DMA_CCR_MEM2MEM | DMA_CCR_MSIZE32 | DMA_CCR_PSIZE32
			| DMA_CCR_MINC | DMA_CCR_EN;
	}
}
